from collections import defaultdict
from typing import Dict, List, Tuple
import sys
import json
import termcolor
import os

def color_by_letter(word, yellows, greens):
    yellow_set = set()
    for m in yellows:
        if m is not None:
            yellow_set.update(m)
    st = ""
    for i, letter in enumerate(word):
        if letter == greens[i]:
            st += termcolor.colored(letter, "green", attrs=['reverse'])
        elif letter in yellow_set:
            st += termcolor.colored(letter, "yellow", attrs=['reverse'])
        else:
            st += letter
    return st
            
    
def read_words(num_letters) -> Tuple[List[str], Dict[str,int]]:
    if not os.path.exists("removed_words"):
        with open("removed_words", "w") as f:
            pass
    with open("removed_words", "r") as f:
        removed_words = {word.strip() for word in f.readlines() if len(word.strip()) == num_letters}

    wf = json.load(open("wf.json", "r"))
    out = []
    words = sorted(wf.items(), key=lambda xs: xs[1], reverse=True)
    for word,_ in words:
        if len(out) >= 5000:
            return out,wf
        if len(word) == num_letters and word not in removed_words:
            out.append(word)
    return out, wf


def setup_reverse_lookup(words):
    rev = defaultdict(set)
    for word in words:
        for letter in word:
            rev[letter].add(word)
    return rev

def normalize_word_freq(wf):
    total = sum(wf.values())
    out = {}
    for k,v in wf.items():
        out[k] = v/total
    return out

def grab_valid(valid_list, wf):
    out = {}
    for word in valid_list:
        out[word] = wf[word]
    return out
    
def find_word(valid_list, rev, yellows, greens, wf, alpha=0.0, beta=1.0, debug = False):
    max_score = 0
    max_word = valid_list[0]
    wf = grab_valid(valid_list, wf)
    wf = normalize_word_freq(wf)
    yellow_set = set()
    for m in yellows:
        if m is not None:
            yellow_set.update(m)
    for word in valid_list:
        s = {word}
        for letter in word:
            if letter not in greens and letter not in yellow_set:
                s.update(rev[letter])
        weighted_count = 0
        for inner_word in s:
            weighted_count += (1-alpha)*1 + alpha*wf[inner_word]
        score = weighted_count * ((1-beta)*1 + beta*wf[word])
        if len(valid_list) <= 50 and debug:
            print(f"{word}: {score}")
        if score > max_score:
            max_score = score
            max_word = word
    if len(valid_list) <= 50 and debug:
        print()
    return max_word

def new_valid(word_list,wf, yellows, greens, greys):
    xs = []
    for word in word_list:
        add = True
        for letter in word:
            if letter in greys:
                add = False
        for letter,expected in zip(word,greens):
            if expected is not None:
                if letter != expected:
                    add = False
                    break
        for letter,yellow in zip(word, yellows):
            if letter in yellow:
                add = False
                break
        yellow_set = set()
        for m in yellows:
            if m is not None:
                yellow_set.update(m)
        if not (yellow_set < set(word)):
            add = False
        if add:
            xs.append(word)
    return xs


def parse_string(word, words:List[str], st, yellows, greens, greys):
    if st[0] == "z":
        words.remove(word)
        with open("removed_words", "a") as f:
            f.write(word+"\n")
        return "z"
    if set(st) == set("g"):
        print("Congrats!")
        return "g"
    for i,(letter,type) in enumerate(zip(word,st)):
        yellow_set = set()
        for m in yellows:
            if m is not None:
                yellow_set.update(m)
        if type == "g":
            greens[i] = letter
        if type == "y":
            yellows[i].add(letter)
        if type == "r":
            if letter in yellow_set or letter in greens:
                yellows[i].add(letter)
            else:
                greys.add(letter)

INSTRUCTIONS = f"""
A solver for https://www.powerlanguage.co.uk/wordle/ using maximum word set differentiation with scheduled english frequency weighting.

USAGE: {sys.argv[0]} [--debug] num_letters

Each iteration the solver will print out the word it thinks is the best next guess. You then input what the website returned (g for green, y for yellow, r for grey) and it'll give its next guess.
For example:
Let's say the solver output
 
  noise>

for the first guess. You should input noise into the website.
Let's assume the website returned
  {color_by_letter("noise", [{}, {"o"}, {}, {}, {}], [None, None, None, None, "e"])}
on that guess.
you would then type ryrrg like
 
  noise> ryrrg
 
then hit enter and the solver will return its next guess. Continue the loop until the solver returns the correct answer, in which case, just type ggggg and it'll end that round.

If the solver gives a word that the website considers invalid, just type z like

  rosie> z

and it'll try another word (and ban that word from being used in future runs).
"""
def main():
    alpha_schedule = [1.0, 0.5, 0.1] + [0.0]*10000
    beta_schedule = [0.0, 0.2,0.5] + [1.0]*10000
    debug = False
    num_letters = 5
    if len(sys.argv) == 1:
        print(INSTRUCTIONS)
        sys.exit(-1)
    elif len(sys.argv) == 2:
        if sys.argv[1] == "--debug":
            debug = True
        else:
            num_letters = int(sys.argv[1])
    elif len(sys.argv) == 3:
        if sys.argv[1] == "--debug":
            debug = True
            num_letters == int(sys.argv[2])
        elif sys.argv[2] == "--debug":
            debug = True
            num_letters = int(sys.argv[1])
        else:
            print(INSTRUCTIONS)
            sys.exit(-1)
    else:
        print(INSTRUCTIONS)
        sys.exit(-1)
            
    while True:
        count = 0
        orig_words,orig_wf = read_words(num_letters)
        words = orig_words
        wf = dict(orig_wf)
        yellows =[set() for _ in range(num_letters)]
        greens = [None for _ in range (num_letters)]
        greys = set()
        while True:
            rev = setup_reverse_lookup(words)
            word = find_word(words, rev, yellows, greens, wf, alpha_schedule[count], beta_schedule[count], debug)
            inp = input(f"{word}> ")
            res = parse_string(word, words, inp, yellows, greens, greys) 
            if res == "g":
                break
            elif res == "z":
                count -= 1
            words = new_valid(words,wf, yellows, greens, greys)
            count += 1
    

if __name__ == "__main__":
    main()
